#include "qusbinfo.h"
#include "qusbinfo_p.h"
#include <QThread>

#define DbgPrintError() qWarning("In %s, at %s:%d", Q_FUNC_INFO, __FILE__, __LINE__)
#define DbgPrintFuncName()                   \
    if (m_log_level >= QUsbDevice::logDebug) \
    qDebug() << "***[" << Q_FUNC_INFO << "]***"
#define DbgPrintCB()                              \
    if (info->logLevel() >= QUsbDevice::logDebug) \
    qDebug() << "***[" << Q_FUNC_INFO << "]***"

static libusb_hotplug_callback_handle callback_handle;

static int LIBUSB_CALL hotplugCallback(libusb_context *ctx,
                                       libusb_device *device,
                                       libusb_hotplug_event event,
                                       void *user_data)
{

    static libusb_device_handle *handle = Q_NULLPTR;
    struct libusb_device_descriptor desc;
    (void)ctx;
    int rc = 0;
    QUsbDevice::IdList device_list;
    QUsbDevice::Id dev;
    QUsbInfo *info = reinterpret_cast<QUsbInfo *>(user_data);
    DbgPrintCB();

    if (info->logLevel())
        qDebug("hotplugCallback");

    (void)libusb_get_device_descriptor(device, &desc);
    if (event == LIBUSB_HOTPLUG_EVENT_DEVICE_ARRIVED) {
        rc = libusb_open(device, &handle);
        if (rc != LIBUSB_SUCCESS) {
            qWarning("Could not open new USB device");
            return rc;
        }
        // Add to list
        dev.vid = desc.idVendor;
        dev.pid = desc.idProduct;
        device_list.append(dev);
        emit info->deviceInserted(device_list);

    } else if (event == LIBUSB_HOTPLUG_EVENT_DEVICE_LEFT) {
        if (handle) {
            // Remove from list
            dev.vid = desc.idVendor;
            dev.pid = desc.idProduct;
            device_list.append(dev);
            emit info->deviceRemoved(device_list);

            libusb_close(handle);
            handle = Q_NULLPTR;
            return 0;
        }
    } else {
        qWarning("Unhandled hotplug event %d", event);
        return -1;
    }
    return rc;
}

QUsbInfoPrivate::QUsbInfoPrivate()
    : m_has_hotplug(false), m_ctx(Q_NULLPTR), m_refresh_timer(new QTimer)
{
    QThread *t = new QThread();
    m_refresh_timer->moveToThread(t);

    m_refresh_timer->setSingleShot(false);
    m_refresh_timer->setInterval(250);

    m_refresh_timer->connect(t, SIGNAL(started()), SLOT(start()));
    m_refresh_timer->connect(t, SIGNAL(finished()), SLOT(stop()));
    t->start();
}

QUsbInfoPrivate::~QUsbInfoPrivate()
{
    m_refresh_timer->thread()->exit();
    m_refresh_timer->thread()->wait();
    m_refresh_timer->thread()->deleteLater();

    m_refresh_timer->disconnect();
    m_refresh_timer->deleteLater();
}

/*!
    \class QUsbInfo

    \brief This class handles hotplug and device detection.

    \reentrant
    \ingroup usb-main
    \inmodule QtUsb

    Handles USB events and searching.
    Can be used to monitor events for a list of devices or all system devices.

    \sa QUsbDevice
*/

/*!
    \fn void QUsbInfo::deviceInserted(QUsbDevice::IdList list)

    This is signal is emited when one or more new devices are detected, providing a \a list.
*/

/*!
    \fn void QUsbInfo::deviceRemoved(QUsbDevice::IdList list)

    This is signal is emited when one or more new devices are removed, providing a \a list.
*/

/*!
    \property QUsbInfo::logLevel
    \brief the log level for hotplug/detection.
*/

QUsbInfo::QUsbInfo(QObject *parent)
    : QObject(*(new QUsbInfoPrivate), parent), d_dummy(Q_NULLPTR)
{
    Q_D(QUsbInfo);

    m_log_level = QUsbDevice::logInfo;
    DbgPrintFuncName();
    int rc;

    qRegisterMetaType<QUsbDevice::Id>("QUsbDevice::DeviceFilter");
    qRegisterMetaType<QUsbDevice::Config>("QUsbDevice::DeviceConfig");

    qRegisterMetaType<QUsbDevice::IdList>("QUsbDevice::FilterList");
    qRegisterMetaType<QUsbDevice::ConfigList>("QUsbDevice::ConfigList");

    rc = libusb_init(&d->m_ctx);
    if (rc < 0) {
        libusb_exit(d->m_ctx);
        qCritical("LibUsb Init Error %d", rc);
        return;
    }

    libusb_set_debug(d->m_ctx, LIBUSB_LOG_LEVEL_WARNING);

    // Populate list once
    m_system_list = QUsbDevice::devices();

    // Try hotplug first
    d->m_has_hotplug = libusb_has_capability(LIBUSB_CAP_HAS_HOTPLUG) != 0;
    if (d->m_has_hotplug) {

        rc = libusb_hotplug_register_callback(d->m_ctx,
                                              static_cast<libusb_hotplug_event>((LIBUSB_HOTPLUG_EVENT_DEVICE_ARRIVED | LIBUSB_HOTPLUG_EVENT_DEVICE_LEFT)),
                                              static_cast<libusb_hotplug_flag>(0),
                                              LIBUSB_HOTPLUG_MATCH_ANY,
                                              LIBUSB_HOTPLUG_MATCH_ANY,
                                              LIBUSB_HOTPLUG_MATCH_ANY,
                                              reinterpret_cast<libusb_hotplug_callback_fn>(hotplugCallback),
                                              reinterpret_cast<void *>(this),
                                              &callback_handle);
        if (LIBUSB_SUCCESS != rc) {
            libusb_exit(d->m_ctx);
            qWarning("Error creating hotplug callback");
            return;
        }
    }

    connect(d->m_refresh_timer, SIGNAL(timeout()), this, SLOT(checkDevices()));
}

/*!
    Unregister callbacks and close the usb context.
 */
QUsbInfo::~QUsbInfo()
{
    Q_D(QUsbInfo);
    DbgPrintFuncName();
    libusb_hotplug_deregister_callback(d->m_ctx, callback_handle);
    libusb_exit(d->m_ctx);
}

/*!
    Check devices present in system.

    This gets called by the internal timer.
 */
void QUsbInfo::checkDevices()
{
    DbgPrintFuncName();
    Q_D(QUsbInfo);
    QUsbDevice::IdList list;

    timeval t = { 0, 0 };

    if (d->m_has_hotplug) {
        libusb_handle_events_timeout_completed(d->m_ctx, &t, Q_NULLPTR);
    } else {
        list = QUsbDevice::devices();
        monitorDevices(list);
    }
}

/*!
   Gets a list of devices present in the system.

 */
QUsbDevice::IdList QUsbInfo::getPresentDevices() const
{
    DbgPrintFuncName();
    QUsbDevice::IdList list;
    QUsbDevice::Id filter;

    /* Search the system list with our own list */
    for (int i = 0; i < m_list.length(); i++) {
        filter = m_list.at(i);
        if (this->findDevice(filter, m_system_list) < 0) {
            list.append(filter);
        }
    }
    return list;
}

/*!
      Check if \a id  device is present.

      Return bool true if present.
 */
bool QUsbInfo::isPresent(const QUsbDevice::Id &id) const
{
    DbgPrintFuncName();
    return this->findDevice(id, m_system_list) >= 0;
}

/*!
      Add an \a id device to the list.

      Returns false if device was already in the list, else true.
 */
bool QUsbInfo::addDevice(const QUsbDevice::Id &id)
{

    DbgPrintFuncName();
    if (this->findDevice(id, m_list) == -1) {
        m_list.append(id);
        return true;
    }
    return false;
}

/*!
      \brief Remove \a id device from the list.

      Return bool false if device was not in the list, else true.
 */
bool QUsbInfo::removeDevice(const QUsbDevice::Id &id)
{

    DbgPrintFuncName();
    const int pos = this->findDevice(id, m_list);
    if (pos > 0) {
        m_list.removeAt(pos);
        return true;
    }
    return true;
}

/*!
      Search an \a id device in a device \a list.

      Return index of the filter, returns -1 if not found.
 */
int QUsbInfo::findDevice(const QUsbDevice::Id &id,
                         const QUsbDevice::IdList &list) const
{
    DbgPrintFuncName();
    for (int i = 0; i < list.length(); i++) {
        const QUsbDevice::Id *d = &list.at(i);

        if (d->pid == id.pid && d->vid == id.vid) {
            return i;
        }
    }
    return -1;
}

/*!
    Set log \a level (only hotplug/detection).
 */
void QUsbInfo::setLogLevel(QUsbDevice::LogLevel level)
{
    DbgPrintFuncName();
    Q_D(QUsbInfo);
    m_log_level = level;
    if (m_log_level >= QUsbDevice::logDebug)
        libusb_set_debug(d->m_ctx, LIBUSB_LOG_LEVEL_DEBUG);
    else
        libusb_set_debug(d->m_ctx, LIBUSB_LOG_LEVEL_WARNING);
}

/*!
    Get current log level.
 */
QUsbDevice::LogLevel QUsbInfo::logLevel() const
{
    return m_log_level;
}

/*!
    Add a \a list to monitor.
 */
void QUsbInfo::monitorDevices(const QUsbDevice::IdList &list)
{

    DbgPrintFuncName();
    QUsbDevice::IdList inserted, removed;
    QUsbDevice::Id filter;

    for (int i = 0; i < list.length(); i++) {
        filter = list.at(i);
        if (this->findDevice(filter, m_system_list) < 0) {
            // It's not in the old system list
            inserted.append(filter);
        }
    }

    for (int i = 0; i < m_system_list.length(); i++) {
        filter = m_system_list.at(i);
        if (this->findDevice(filter, list) < 0) {
            // It's in the old system list but not in the current one
            removed.append(filter);
        }
    }

    if (inserted.length() > 0)
        emit deviceInserted(inserted);

    if (removed.length() > 0)
        emit deviceRemoved(removed);

    m_system_list = list;
}
